/* Copyright (C) 2016-2021 |Meso|Star> (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s2d.h"
#include "test_s2d_utils.h"

#include <rsys/float2.h>

#define NSAMPS 4096

int
main(int argc, char** argv)
{
  struct mem_allocator allocator;
  struct s2d_vertex_data vdata;
  struct s2d_device* dev;
  struct s2d_scene* scn;
  struct s2d_scene_view* scnview;
  struct s2d_shape* shape;
  struct s2d_primitive prim;
  float s;
  unsigned square_id;
  int i;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHK(s2d_device_create(NULL, &allocator, 1, &dev) == RES_OK);
  CHK(s2d_scene_create(dev, &scn) == RES_OK);
  CHK(s2d_shape_create_line_segments(dev, &shape) == RES_OK);

  CHK(s2d_shape_get_id(shape, &square_id) == RES_OK);
  CHK(s2d_scene_attach_shape(scn, shape) == RES_OK);
  CHK(s2d_scene_view_create(scn, S2D_SAMPLE, &scnview) == RES_OK);

  CHK(s2d_scene_view_sample(NULL, 1, 1, NULL, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 1, 1, NULL, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 0, 1, NULL, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 0, 1, NULL, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 1, 0, NULL, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 1, 0, NULL, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 0, 0, NULL, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 0, 0, NULL, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 1, 1, &prim, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 1, 1, &prim, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 0, 1, &prim, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 0, 1, &prim, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 1, 0, &prim, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 1, 0, &prim, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 0, 0, &prim, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 0, 0, &prim, NULL) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 1, 1, NULL, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 1, 1, NULL, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 0, 1, NULL,&s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 0, 1, NULL, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 1, 0, NULL, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 1, 0, NULL, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 0, 0, NULL, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 0, 0, NULL, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 1, 1, &prim, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 1, 1, &prim, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 0, 1, &prim, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 0, 1, &prim, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 1, 0, &prim, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 1, 0, &prim, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(NULL, 0, 0, &prim, &s) == RES_BAD_ARG);
  CHK(s2d_scene_view_sample(scnview, 0, 0, &prim, &s) == RES_OK);
  CHK(S2D_PRIMITIVE_EQ(&prim, &S2D_PRIMITIVE_NULL) == 1);

  vdata.usage = S2D_POSITION;
  vdata.type = S2D_FLOAT2;
  vdata.get = line_segments_get_position;
  CHK(s2d_line_segments_setup_indexed_vertices
    (shape, square_nsegs, line_segments_get_ids, square_nverts, &vdata, 1,
     (void*)&square_desc) == RES_OK);

  CHK(s2d_scene_view_sample(scnview, 0, 0, &prim, &s) == RES_OK);
  CHK(S2D_PRIMITIVE_EQ(&prim, &S2D_PRIMITIVE_NULL) == 1);

  CHK(s2d_scene_view_ref_put(scnview) == RES_OK);

  CHK(s2d_scene_view_create(scn, S2D_SAMPLE, &scnview) == RES_OK);
  CHK(s2d_scene_view_sample(scnview, 0, 0, &prim, &s) == RES_OK);
  CHK(S2D_PRIMITIVE_EQ(&prim, &S2D_PRIMITIVE_NULL) == 0);

  CHK(prim.prim_id < 4);
  CHK(prim.geom_id == square_id);
  CHK(s == 0);

  /* Should take effect int the next created scene view */
  CHK(s2d_shape_flip_contour(shape) == RES_OK);

  FOR_EACH(i, 0, NSAMPS) {
    struct s2d_attrib attr_position;
    struct s2d_attrib attr_normal;
    const float u = rand_canonic();
    const float v = rand_canonic();
    float N[2], P[2], tmp[2];

    CHK(s2d_scene_view_sample(scnview, u, v, &prim, &s) == RES_OK);
    CHK(S2D_PRIMITIVE_EQ(&prim, &S2D_PRIMITIVE_NULL) == 0);
    CHK(prim.prim_id < 4);
    CHK(prim.geom_id == square_id);

    CHK(s2d_primitive_get_attrib
      (&prim, S2D_POSITION, s, &attr_position) == RES_OK);
    CHK(s2d_primitive_get_attrib
      (&prim, S2D_GEOMETRY_NORMAL, s, &attr_normal) == RES_OK);

    f2_normalize(attr_normal.value, attr_normal.value);

    switch(prim.prim_id) {
      case 0: f2(P, -1.f*s + (1-s)*1.f, -1.f); f2(N, 0.f, 1.f); break;
      case 1: f2(P, -1.f, 1.f*s + (1-s)*-1.f); f2(N, 1.f, 0.f); break;
      case 2: f2(P, 1.f*s + (1-s)* -1.f, 1.f); f2(N, 0.f,-1.f); break;
      case 3: f2(P, 1.f, -1.f*s + (1-s)* 1.f); f2(N,-1.f, 0.f); break;
      default: CHK(0 == 1); break; /* Invalid primitive id */
    }
    f2_add(P, P, f2_splat(tmp, 10));
    CHK(f2_eq_eps(P, attr_position.value, 1.e-6f));
    CHK(f2_eq_eps(N, attr_normal.value, 1.e-6f));
  }

  CHK(s2d_scene_view_ref_put(scnview) == RES_OK);

  CHK(s2d_scene_view_create(scn, S2D_SAMPLE, &scnview) == RES_OK);
  FOR_EACH(i, 0, NSAMPS) {
    struct s2d_attrib attr;
    const float u = rand_canonic();
    const float v = rand_canonic();
    float N[2];

    CHK(s2d_scene_view_sample(scnview, u, v, &prim, &s) == RES_OK);
    CHK(S2D_PRIMITIVE_EQ(&prim, &S2D_PRIMITIVE_NULL) == 0);
    CHK(prim.prim_id < 4);
    CHK(prim.geom_id == square_id);

    CHK(s2d_primitive_get_attrib(&prim, S2D_GEOMETRY_NORMAL, s, &attr) == RES_OK);
    f2_normalize(attr.value, attr.value);

    switch(prim.prim_id) {
      case 0: f2(N, 0.f,-1.f); break;
      case 1: f2(N,-1.f, 0.f); break;
      case 2: f2(N, 0.f, 1.f); break;
      case 3: f2(N, 1.f, 0.f); break;
      default: CHK(0 == 1); break; /* Invalid primitive id */
    }

    CHK(f2_eq_eps(N, attr.value, 1.e-6f) == 1);
  }

  CHK(s2d_scene_view_ref_put(scnview) == RES_OK);

  CHK(s2d_device_ref_put(dev) == RES_OK);
  CHK(s2d_scene_ref_put(scn) == RES_OK);
  CHK(s2d_shape_ref_put(shape) == RES_OK);

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHK(mem_allocated_size() == 0);
  return 0;
}

