/* Copyright (C) 2016-2021 |Meso|Star> (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s2d.h"
#include "test_s2d_utils.h"

#include <rsys/float2.h>
#include <rsys/float3.h>

int
main(int argc, char** argv)
{
  struct mem_allocator allocator;
  struct s2d_device* dev;
  struct s2d_scene* scn;
  struct s2d_scene_view* scnview;
  struct s2d_shape* shape;
  struct s2d_vertex_data vdata;
  struct s2d_hit hit;
  float org[3];
  float dir[3];
  float range[2];
  float tmp[2];
  float dot;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHK(s2d_device_create(NULL, &allocator, 1, &dev) == RES_OK);
  CHK(s2d_scene_create(dev, &scn) == RES_OK);
  CHK(s2d_shape_create_line_segments(dev, &shape) == RES_OK);
  CHK(s2d_scene_attach_shape(scn, shape) == RES_OK);

  vdata.get = line_segments_get_position;
  vdata.type = S2D_FLOAT2;
  vdata.usage = S2D_POSITION;
  CHK(s2d_line_segments_setup_indexed_vertices
    (shape, square_nsegs, line_segments_get_ids, square_nverts, &vdata, 1,
     (void*)&square_desc) == RES_OK);

  f3_splat(org, 10.f);
  f3(dir, 1, 0, 0);
  f2(range, 0, FLT_MAX);

  CHK(s2d_scene_view_create(scn, S2D_TRACE, &scnview) == RES_OK);

  #define RT_3D s2d_scene_view_trace_ray_3d
  CHK(RT_3D(NULL, NULL, NULL, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(scnview, NULL, NULL, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(NULL, org, NULL, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(scnview, org, NULL, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(NULL, NULL, dir, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(scnview, NULL, dir, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(NULL, org, dir, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(scnview, org, dir, NULL, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(NULL, NULL, NULL, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(scnview, NULL, NULL, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(NULL, org, NULL, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(scnview, org, NULL, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(NULL, NULL, dir, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(scnview, NULL, dir, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(NULL, org, dir, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(scnview, org, dir, range, NULL, NULL) == RES_BAD_ARG);
  CHK(RT_3D(NULL, NULL, NULL, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(scnview, NULL, NULL, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(NULL, org, NULL, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(scnview, org, NULL, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(NULL, NULL, dir, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(scnview, NULL, dir, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(NULL, org, dir, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(scnview, org, dir, NULL, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(NULL, NULL, NULL, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(scnview, NULL, NULL, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(NULL, org, NULL, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(scnview, org, NULL, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(NULL, NULL, dir, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(scnview, NULL, dir, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(NULL, org, dir, range, NULL, &hit) == RES_BAD_ARG);
  CHK(RT_3D(scnview, org, dir, range, NULL, &hit) == RES_OK);

  CHK(S2D_HIT_NONE(&hit) == 0);
  f2_normalize(hit.normal, hit.normal);
  CHK(f2_eq_eps(hit.normal, f2(tmp, -1.f, 0.f), 1.e-6f) == 1);
  CHK(eq_epsf(hit.distance, 1.f, 1.e-6f) == 1);
  CHK(eq_epsf(hit.u, 0.5f, 1.e-6f) == 1);

  f3_normalize(dir, f3(dir, 1.f, 0.f, 1.f));
  CHK(RT_3D(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 0);
  f2_normalize(hit.normal, hit.normal);
  CHK(f2_eq_eps(hit.normal, f2(tmp, -1.f, 0.f), 1.e-6f) == 1);
  CHK(eq_epsf(hit.u, 0.5f, 1.e-6f) == 1);

  dot = dir[0];
  CHK(eq_epsf(hit.distance, 1.f / dot, 1.e-6f) == 1);

  range[1] = 1.2f;
  CHK(RT_3D(scnview, org, dir, range, NULL, &hit) == RES_OK);
  CHK(S2D_HIT_NONE(&hit) == 1);

  CHK(s2d_scene_view_ref_put(scnview) == RES_OK);

  CHK(RT_3D(scnview, org, dir, range, NULL, &hit) == RES_BAD_OP);
  #undef RT_3D

  CHK(s2d_scene_ref_put(scn) == RES_OK);
  CHK(s2d_shape_ref_put(shape) == RES_OK);
  CHK(s2d_device_ref_put(dev) == RES_OK);

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHK(mem_allocated_size() == 0);

  return 0;
}
