/* Copyright (C) 2016-2021 |Meso|Star> (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

/* This test has been created using the sg3_geometry_dump_as_C_code feature
 * of star-geometry. It uses output from test_sg3_cube_on_cube. */

#define _POSIX_C_SOURCE 200112L /* snprintf */

#include "senc3d.h"
#include "test_senc3d_utils.h"

#include <rsys/double3.h>

#include <stdio.h>

/* Tests created using -c option of stardis */

/* zero_1 geometry is made of a big distorded cube, that as a single highest
 * vertex V, and two internal shapes (highly distorded cubes) that joint at the
 * same V vertex. Distance between the 6 components (inside and outside of each
 * shape) is zero. */
#define zero_1_UNSPECIFIED_PROPERTY 4294967295

static const unsigned
zero_1_vertices_count = 22;

static const unsigned
zero_1_triangles_count = 36;

static const double
zero_1_vertices[22][3] = {
   { 10, 10, 0 },
   { 10, 0, 0 },
   { 0, 10, 0 },
   { 0, 0, 0 },
   { 0, 10, 10 },
   { 0, 0, 10 },
   { 10, 0, 10 },
   { 9, 9, 11 },
   { 8, 9, 1 },
   { 4, 7, 1 },
   { 1, 9, 1 },
   { 1, 2, 1 },
   { 1, 9, 9 },
   { 1, 2, 9 },
   { 4, 7, 9 },
   { 7, 5, 1 },
   { 9, 8, 1 },
   { 9, 1, 1 },
   { 2, 1, 1 },
   { 2, 1, 9 },
   { 9, 1, 9 },
   { 7, 4, 9 }
};

static const unsigned
zero_1_triangles[36][3] = {
   { 0, 1, 2 },
   { 1, 3, 2 },
   { 4, 5, 6 },
   { 7, 4, 6 },
   { 2, 3, 4 },
   { 3, 5, 4 },
   { 6, 1, 7 },
   { 1, 0, 7 },
   { 0, 2, 4 },
   { 7, 0, 4 },
   { 5, 3, 1 },
   { 6, 5, 1 },
   { 8, 9, 10 },
   { 9, 11, 10 },
   { 12, 13, 14 },
   { 7, 12, 14 },
   { 10, 11, 12 },
   { 11, 13, 12 },
   { 14, 9, 7 },
   { 9, 8, 7 },
   { 8, 10, 12 },
   { 7, 8, 12 },
   { 13, 11, 9 },
   { 14, 13, 9 },
   { 15, 16, 17 },
   { 18, 15, 17 },
   { 19, 20, 21 },
   { 20, 7, 21 },
   { 18, 17, 20 },
   { 19, 18, 20 },
   { 15, 21, 7 },
   { 16, 15, 7 },
   { 17, 16, 20 },
   { 16, 7, 20 },
   { 18, 19, 15 },
   { 19, 21, 15 }
};

static const unsigned
zero_1_properties[36][3] = {
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 0, zero_1_UNSPECIFIED_PROPERTY, 2 },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_1_UNSPECIFIED_PROPERTY }
};

/* zero_2 geometry is made of 5 nested distorded cubes of decreasing sizes, that
 * all join at a single vertex V (the highest vertex in the model). Distance between
 * the 10 components (inside and outside of each shape) is zero. */
#define zero_2_UNSPECIFIED_PROPERTY 4294967295

static const unsigned
zero_2_vertices_count = 36;

static const unsigned
zero_2_triangles_count = 60;

static const double
zero_2_vertices[36][3] = {
   { 0, 10, 0 },
   { 0, 0, 0 },
   { 0, 10, 10 },
   { 0, 0, 10 },
   { 10, 0, 10 },
   { 10, 0, 0 },
   { 9, 9, 11 },
   { 10, 10, 0 },
   { 1, 9, 1 },
   { 1, 1, 1 },
   { 1, 9, 9 },
   { 1, 1, 9 },
   { 9, 1, 9 },
   { 9, 1, 1 },
   { 9, 9, 1 },
   { 2, 8, 2 },
   { 2, 2, 2 },
   { 2, 8, 8 },
   { 2, 2, 8 },
   { 8, 2, 8 },
   { 8, 2, 2 },
   { 8, 8, 2 },
   { 3, 7, 3 },
   { 3, 3, 3 },
   { 3, 7, 7 },
   { 3, 3, 7 },
   { 7, 3, 7 },
   { 7, 3, 3 },
   { 7, 7, 3 },
   { 4, 6, 4 },
   { 4, 4, 4 },
   { 4, 6, 6 },
   { 4, 4, 6 },
   { 6, 4, 6 },
   { 6, 4, 4 },
   { 6, 6, 4 }
};

static const unsigned
zero_2_triangles[60][3] = {
   { 0, 1, 2 },
   { 1, 3, 2 },
   { 4, 5, 6 },
   { 5, 7, 6 },
   { 2, 3, 4 },
   { 6, 2, 4 },
   { 5, 1, 0 },
   { 7, 5, 0 },
   { 3, 1, 5 },
   { 4, 3, 5 },
   { 7, 0, 2 },
   { 6, 7, 2 },
   { 8, 9, 10 },
   { 9, 11, 10 },
   { 12, 13, 6 },
   { 13, 14, 6 },
   { 10, 11, 12 },
   { 6, 10, 12 },
   { 13, 9, 8 },
   { 14, 13, 8 },
   { 11, 9, 13 },
   { 12, 11, 13 },
   { 14, 8, 10 },
   { 6, 14, 10 },
   { 15, 16, 17 },
   { 16, 18, 17 },
   { 19, 20, 6 },
   { 20, 21, 6 },
   { 17, 18, 19 },
   { 6, 17, 19 },
   { 20, 16, 15 },
   { 21, 20, 15 },
   { 18, 16, 20 },
   { 19, 18, 20 },
   { 21, 15, 17 },
   { 6, 21, 17 },
   { 22, 23, 24 },
   { 23, 25, 24 },
   { 26, 27, 6 },
   { 27, 28, 6 },
   { 24, 25, 26 },
   { 6, 24, 26 },
   { 27, 23, 22 },
   { 28, 27, 22 },
   { 25, 23, 27 },
   { 26, 25, 27 },
   { 28, 22, 24 },
   { 6, 28, 24 },
   { 29, 30, 31 },
   { 30, 32, 31 },
   { 33, 34, 6 },
   { 34, 35, 6 },
   { 31, 32, 33 },
   { 6, 31, 33 },
   { 34, 30, 29 },
   { 35, 34, 29 },
   { 32, 30, 34 },
   { 33, 32, 34 },
   { 35, 29, 31 },
   { 6, 35, 31 }
};

static const unsigned
zero_2_properties[60][3] = {
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 0, zero_2_UNSPECIFIED_PROPERTY, 5 },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 1, 0, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 2, 1, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 3, 2, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY },
   { 4, 3, zero_2_UNSPECIFIED_PROPERTY }
};

/* Enclosures' order is consistent accross runs as they are sorted by triangle
 * sides, thus allowing to expect a given order */
struct expected {
  unsigned enclosure_count; /* max 16 */
  double volume[16]; /* the volume of the expected enclosures */
  unsigned medium[16]; /* the single medium of the expected enclosures */
};

static int
test
  (const double* positions,
   const unsigned* indices,
   const unsigned* properties,
   const unsigned triangles_count,
   const unsigned vertices_count,
   const struct expected* expected)
{
  struct mem_allocator allocator;
  struct senc3d_device* dev = NULL;
  struct senc3d_scene* scn = NULL;
  struct context ctx = CONTEXT_NULL__;
  unsigned count, e, n;

  OK(mem_init_proxy_allocator(&allocator, &mem_default_allocator));
  OK(senc3d_device_create(NULL, &allocator, SENC3D_NTHREADS_DEFAULT, 1, &dev));

  FOR_EACH(n, 0, 2) {
    /* Create a scene with 4 enclosures joining at a single vertex. */
    ctx.positions = positions;
    ctx.indices = indices;
    ctx.properties = properties;

    if(n == 1) {
      /* Same model with symmetry / origin
       * (need to reverse vertices' order to keep same inside/outside) */
      ctx.scale = -1;
      ctx.reverse_vrtx = 1;
    }

    OK(senc3d_scene_create(dev,
      SENC3D_CONVENTION_NORMAL_BACK | SENC3D_CONVENTION_NORMAL_OUTSIDE,
      triangles_count, get_indices, get_media_from_properties,
      vertices_count, get_position, &ctx, &scn));

    OK(senc3d_scene_get_vertices_count(scn, &count));
    CHK(count == vertices_count);

    OK(senc3d_scene_get_triangles_count(scn, &count));
    CHK(count == triangles_count);

    OK(senc3d_scene_get_enclosure_count(scn, &count));
    CHK(count == expected->enclosure_count);

    FOR_EACH(e, 0, count) {
      struct senc3d_enclosure* enclosure;
      struct senc3d_enclosure_header header;
      unsigned m;
      char name[128]; (void)name;
      OK(senc3d_scene_get_enclosure(scn, e, &enclosure));
      OK(senc3d_enclosure_get_header(enclosure, &header));
      CHK(header.enclosed_media_count == 1);
      OK(senc3d_enclosure_get_medium(enclosure, 0, &m));
      CHK(m == expected->medium[e]);
      CHK(eq_eps(header.volume, expected->volume[e], 0.01));
      OK(senc3d_enclosure_ref_put(enclosure));
#ifdef DUMP_ENCLOSURES
      snprintf(name, sizeof(name), "test_cube_on_cube_%u.obj", e);
      dump_enclosure(scn, e, name);
#endif
    }

    OK(senc3d_scene_ref_put(scn));
  }

  OK(senc3d_device_ref_put(dev));

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHK(mem_allocated_size() == 0);
  return 0;
}

int
main(int argc, char** argv)
{
  struct expected expected;
  (void)argc, (void)argv;

  expected.enclosure_count = 4;
  expected.medium[0] = SENC3D_UNSPECIFIED_MEDIUM;
  expected.volume[0] = -966.67;
  expected.medium[1] = 0;
  expected.volume[1] = 647.33;
  expected.medium[2] = 1;
  expected.volume[2] = 150;
  expected.medium[3] = 1;
  expected.volume[3] = 169.33;
  test((double *)zero_1_vertices, (unsigned*)zero_1_triangles,
    (unsigned *)zero_1_properties, zero_1_triangles_count,
    zero_1_vertices_count, &expected);

  expected.enclosure_count = 6;
  expected.medium[0] = SENC3D_UNSPECIFIED_MEDIUM;
  expected.volume[0] = -966.67;
  expected.medium[1] = 0;
  expected.volume[1] = 433.33;
  expected.medium[2] = 1;
  expected.volume[2] = 281.33;
  expected.medium[3] = 2;
  expected.volume[3] = 161.33;
  expected.medium[4] = 3;
  expected.volume[4] = 73.33;
  expected.medium[5] = 4;
  expected.volume[5] = 17.33;
  test((double*)zero_2_vertices, (unsigned*)zero_2_triangles,
    (unsigned*)zero_2_properties, zero_2_triangles_count,
    zero_2_vertices_count, &expected);

  return 0;
}